<?php

/**
 * @file
 * Administration forms for managing form/field overrides.
 */

define('AC_ATTR_ADMIN_SECTIONS_RGID', 1);

/**
 * Form Rules.
 */
function ac_attr_admin_rules($form, &$form_state) {
  drupal_add_js(drupal_get_path('module', 'ac_attr') . '/js/admin_forms.js');
  drupal_add_css(drupal_get_path('module', 'ac_attr') . '/css/admin_forms.css');

  $form['#attributes'] = [
    'class' => ['ac-attr-admin'],
  ];

  $about =
    t('These are the rules used to set the autocomplete attribute on form fields through out the site.<br>Rules work by selecting field paths and applying the preset autofill token on fields that match it.<br>Field path segments are seperated by a dot.<br>A backslash can be used to escape character if need be.<br>Multiple field path patterns can be specified by seperating them with a new line.<br>')
    . theme('item_list', [
      'title' => t('There are several special selectors that can be used when specifing field paths.'),
      'items' => [
        '<span class="special-selector">*</span> ' . t('Select the next one field path segment regardless of what it is.'),
        '<span class="special-selector">**</span> ' . t('Select any number (including zero) of the next lot of field path segments regardless of what they are.'),
      ],
      'type' => 'ul',
      'attributes' => [],
    ])
    . '<br>'
    . t('For more information about the autocomplete attribute, see <a href="@spec_url" aria-label="@spec_url_aria_label" target="_blank">the spec</a>.<br>', [
      '@spec_url' => 'https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill',
      '@spec_url_aria_label' => t('Autocomplete attribute specification.'),
    ]);

  $form['about'] = [
    '#type' => 'markup',
    '#markup' => "<p>$about<p>",
    '#prefix' => '<div class="messages info">',
    '#suffix' => '</div>',
    '#weight' => -10,
  ];

  _ac_attr_admin_add_form_helpers($form);

  $rule_groups = db_select('ac_attr_rule_groups', 'rg')
    ->fields('rg')
    ->execute()
    ->fetchAllAssoc('rgid', PDO::FETCH_ASSOC);

  $rules_by_group_id = _ac_attr_get_rules_by_group_id(TRUE);

  // Sort the sections (if there are any) by title.
  if (array_key_exists(AC_ATTR_ADMIN_SECTIONS_RGID, $rules_by_group_id)) {
    uasort(
      $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID],
      function ($a, $b) {
        return strnatcmp($a['title'], $b['title']);
      }
    );
  }
  // Create an empty sections list if it doesn't exist.
  else {
    $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID] = [];
  }

  $next_new_rule_id = 0;
  _ac_attr_admin_rules_rebuild(
    $form_state,
    $rules_by_group_id,
    $next_new_rule_id
  );

  $form['root_patterns'] = [
    '#type' => 'textarea',
    '#title' => t('Root Field Path Patterns'),
    '#default_value' => variable_get(
      'ac_attr_root_field_path_patterns',
      AC_ATTR_DEFAULT_ROOT_FIELD_PATH_PATTERNS
    ),
    '#required' => TRUE,
    '#rows' => 2,
    '#weight' => 1,
    '#attributes' => [
      'class' => ['field-path-patterns'],
      'autocomplete' => 'off',
    ],
    '#description' => t('These patterns are prefixed to all other rules.'),
  ];

  $form['rules'] = [
    '#tree' => TRUE,
    '#weight' => 5,
  ];

  $rule_group_to_form_group_map = [];
  $unclaimed_children = [];

  // Create form groups for each rule group.
  foreach ($rule_groups as $rule_group) {
    $rgid = (int) $rule_group['rgid'];

    // Create the form group for this group.
    $rule_group_to_form_group_map[$rgid] = [
      '#weight' => $rgid,
    ];

    $group = &$rule_group_to_form_group_map[$rgid];
    if ($rule_group['name'] !== NULL) {
      $group += _ac_attr_admin_rules_get_nested_group_form_element($rule_group['name']);
    }
    $group['rgid'] = [
      '#type' => 'value',
      '#value' => $rgid,
    ];
    $group['rules'] = [
      '#type' => 'fieldset',
      '#title' => t('Autofill Field Tokens'),
      '#collapsible' => TRUE,
      '#weight' => 5,
    ];
    $group['rules']['empty'] = [
      '#type' => 'markup',
      '#weight' => 5,
    ];

    // Does this group have any unclaimed children? If so, claim them.
    if (
      array_key_exists($rgid, $unclaimed_children)
      && count($unclaimed_children[$rgid]) > 0
    ) {
      $group['subrules'] = _ac_attr_admin_rules_get_nested_group_form_element();
      foreach ($unclaimed_children[$rgid] as &$child) {
        $group['subrules'][] = &$child;
      }
      unset($unclaimed_children[$rgid]);
    }

    // Does this group have any rules?
    if (
      array_key_exists($rgid, $rules_by_group_id)
      && count($rules_by_group_id[$rgid]) > 0
    ) {
      // Loop through each of this group's rules.
      foreach ($rules_by_group_id[$rgid] as $rule) {
        $group['rules']['list']['#weight'] = 10;

        $new_rule = isset($rule['new_rule'])
          && $rule['new_rule'];

        $completely_new_rule = $new_rule && empty($rule['value']);

        $rule_id = $new_rule
          ? (isset($rule['new_rule_id'])
            ? $rule['new_rule_id']
            : ('new-' . ($next_new_rule_id++)))
          : 'rid-' . $rule['rid'];

        $group_title = $completely_new_rule
          ? ($rgid === AC_ATTR_ADMIN_SECTIONS_RGID ? t('New Section') : t('New Rule'))
          : (empty($rule['title'])
            ? ($rule['disabled']
              ? t('@ac_value [disabled]', ['@ac_value' => $rule['value']])
              : t('@ac_value', ['@ac_value' => $rule['value']])
            )
            : ($rule['disabled']
              ? t('@title (@ac_value) [disabled]', [
                // phpcs:ignore Drupal.Semantics.FunctionT -- Built in titles need to be translated.
                '@title' => $rule['built_in'] ? t($rule['title']) : $rule['title'],
                '@ac_value' => $rule['value'],
              ])
              : t('@title (@ac_value)', [
                // phpcs:ignore Drupal.Semantics.FunctionT -- Built in titles need to be translated.
                '@title' => $rule['built_in'] ? t($rule['title']) : $rule['title'],
                '@ac_value' => $rule['value'],
              ])
            )
          );

        $group['rules']['list'][$rule_id] = [
          '#type' => 'fieldset',
          '#title' => $group_title,
          // phpcs:ignore Drupal.Semantics.FunctionT -- Built in descriptions need to be translated.
          '#description' => $rule['built_in'] ? t($rule['description']) : $rule['description'],
          '#collapsible' => TRUE,
          '#collapsed' => !$completely_new_rule,
        ];
        $group['rules']['list'][$rule_id]['rid'] = [
          '#type' => 'value',
          '#value' => $rule_id,
        ];
        $group['rules']['list'][$rule_id]['patterns'] = [
          '#type' => 'textarea',
          '#title' => t('Field Path Patterns'),
          '#default_value' => $rule['field_path_patterns'],
          '#attributes' => [
            'placeholder' => $rgid === AC_ATTR_ADMIN_SECTIONS_RGID ? 'section-xyz.**.' : 'foo',
            'class' => ['field-path-patterns'],
            'autocomplete' => 'off',
          ],
          '#rows' => max(2, min(5,
            substr_count($rule['field_path_patterns'], "\n") + 1
          )),
          '#required' => TRUE,
          '#weight' => 10,
          '#description' => t(
            "The patterns to match form field paths against to test whether or not this rule is applied.<br>These patterns will each be prefixed with the patterns of the parent rules."
          ),
        ];
        $group['rules']['list'][$rule_id]['disabled'] = [
          '#type' => 'checkbox',
          '#title' => t('Disbaled'),
          '#default_value' => (bool) $rule['disabled'],
          '#attributes' => [
            'autocomplete' => 'off',
          ],
          '#description' => t('When set, this rule will never be applied.'),
          '#weight' => 20,
        ];
        $group['rules']['list'][$rule_id]['built_in'] = [
          '#type' => 'value',
          '#value' => (bool) $rule['built_in'],
        ];
        if ($rule['built_in']) {
          $group['rules']['list'][$rule_id]['title'] = [
            '#type' => 'value',
            '#value' => $rule['title'],
          ];
          $group['rules']['list'][$rule_id]['value'] = [
            '#type' => 'value',
            '#value' => $rule['value'],
          ];
        }
        else {
          $group['rules']['list'][$rule_id]['title'] = [
            '#type' => 'textfield',
            '#title' => t('Title'),
            '#default_value' => $rule['title'],
            '#attributes' => [
              'placeholder' => $rgid === AC_ATTR_ADMIN_SECTIONS_RGID ? 'My Section' : 'My Rule',
              'autocomplete' => 'off',
            ],
            '#required' => FALSE,
            '#description' => t('The title that is displayed for this rule.'),
            '#weight' => 1,
          ];
          $group['rules']['list'][$rule_id]['value'] = [
            '#type' => 'textfield',
            '#title' => t('Token'),
            '#default_value' => $rule['value'],
            '#attributes' => [
              'placeholder' => $rgid === AC_ATTR_ADMIN_SECTIONS_RGID ? 'section-xyz' : 'foo',
              'autocomplete' => 'off',
            ],
            '#required' => TRUE,
            '#description' => t(
              'The value that will be set on the autocomplete attribute for fields that match this rule.'
            ),
            '#weight' => 12,
          ];
          // Add a button to remove this section.
          $group['rules']['list'][$rule_id]['actions'] = [
            '#type' => 'actions',
            'delete' => [
              '#type' => 'button',
              '#value' => t('Remove'),
              '#name' => "delete-section-$rule_id",
              '#attributes' => [
                'formnovalidate' => TRUE,
              ],
              '#limit_validation_errors' => [],
              // TODO: Add AJAX handler.
              '#weight' => 2,
            ],
            '#weight' => 100,
          ];
        }
      }
    }
    // No rules?
    else {
      $group['rules']['empty']['#prefix'] = '<div class="no-rules">';
      $group['rules']['empty']['#markup'] = t('No rules');
      $group['rules']['empty']['#suffix'] = '</div>';
    }

    // Does this group have a parent?
    if ($rule_group['parent_group'] !== NULL) {
      // If we have already processed the parent, add this fieldset to it.
      if (array_key_exists($rule_group['parent_group'], $rule_group_to_form_group_map)) {
        if (!array_key_exists('subrules', $rule_group_to_form_group_map[$rule_group['parent_group']])) {
          $rule_group_to_form_group_map[$rule_group['parent_group']]['subrules'] = _ac_attr_admin_rules_get_nested_group_form_element();
        }
        $rule_group_to_form_group_map[$rule_group['parent_group']]['subrules'][] = &$group;
      }
      // Otherwise save it so it can be added later to its parnet later.
      else {
        $unclaimed_children[$rule_group['parent_group']][] = &$group;
      }
    }
    // Top-level? Add it to the form.
    else {
      $form['rules']["group-$rgid"] = &$group;
    }
  }

  $sections = &$form['rules']['group-' . AC_ATTR_ADMIN_SECTIONS_RGID];
  $sections['#title'] = empty($rule_groups[AC_ATTR_ADMIN_SECTIONS_RGID]['name'])
    ? t('Sections')
    // phpcs:ignore Drupal.Semantics.FunctionT -- Built in names need to be translated.
    : t($rule_groups[AC_ATTR_ADMIN_SECTIONS_RGID]['name']);
  $sections['#type'] = 'container';
  $sections['#id'] = 'rule-sections';

  $sections['rules']['#title'] = $sections['#title'];
  if (isset($sections['rules']['empty']['#markup'])) {
    $sections['rules']['empty']['#markup'] = t('There are currently no sections defined.');
  }

  // Add a button to add sections.
  $sections['rules']['actions'] = [
    '#type' => 'actions',
    'add' => [
      '#type' => 'button',
      '#value' => t('Add Section'),
      '#name' => 'add_section',
      // TODO: Add AJAX handler.
    ],
    '#weight' => 100,
  ];

  $sections_subrules_count = 0;
  foreach ($sections['subrules'] as $key => &$section_subrules) {
    // Form API field? Skip it.
    if (substr($key, 0, 1) === '#') {
      continue;
    }

    $sections_subrules_count++;
    $section_subrules['#title'] = empty($rule_groups[$section_subrules['rgid']['#value']]['name'])
      ? t('Rule Group')
      : t('Rule Group: @group_name', [
        '@group_name' => $rule_groups[$section_subrules['rgid']['#value']]['name'],
      ]);
  }

  if ($sections_subrules_count > 1) {
    $sections['subrules']['#title'] = t('Rules');
  }
  else {
    $sections['subrules']['#type'] = 'container';

    if ($sections_subrules_count === 1) {
      $sections['subrules'][0]['#title'] = t('Rules');
    }
  }

  // Loop through again now that all the basics are set.
  foreach ($rule_groups as $rule_group) {
    $rgid = (int) $rule_group['rgid'];
    $group = &$rule_group_to_form_group_map[$rgid];
    // Are there subrules?
    if (
      array_key_exists('subrules', $group)
      && is_array($group['subrules'])
      && count($group['subrules']) > 0
    ) {
      // Add the optional checkbox.
      $group['rules']['optional'] = [
        '#type' => 'checkbox',
        '#title' => t('Optional'),
        '#value' => (bool) $rule_group['optional'],
        '#description' => t('When set, these rules do not need to match in order for a nested autofill field token to be applied.'),
        '#disabled' => TRUE,
        '#weight' => 1,
      ];

      if (isset($group['rules']['list'])) {
        // Add the requires subvalue checkbox.
        foreach ($group['rules']['list'] as $list_key => &$rule_list) {
          if (strpos($list_key, '#') === 0) {
            continue;
          }
          $rid = $rule_list["rid"]["#value"];
          $rule = $rules_by_group_id[$rgid][$rid];
          $rule_list['requires_subvalue'] = [
            '#type' => 'checkbox',
            '#title' => t('Requires Subvalue'),
            '#value' => (bool) $rule['requires_subvalue'],
            '#description' => t('When set, this rule is only applied when a nested autofill field token is also applied.'),
            '#disabled' => TRUE,
            '#weight' => 50,
          ];
        }
      }
    }
  }

  // Add the submit button.
  $form['actions'] = [
    '#type' => 'actions',
    'submit' => [
      '#type' => 'submit',
      '#value' => t('Submit'),
    ],
  ];

  $form['#validate'][] = 'ac_attr_admin_rules_validate';

  return $form;
}

/**
 * Restore the ac_attr_admin_rules form if it is being rebuild.
 *
 * @param array $form_state
 *   The form state.
 * @param array $rules_by_group_id
 *   The rules sorted into groups.
 * @param int $next_new_rule_id
 *   The next new rule id to use.
 */
function _ac_attr_admin_rules_rebuild(
  array &$form_state,
  array &$rules_by_group_id,
  &$next_new_rule_id
) {
  // If the form is not being rebuilt, get out of here.
  if (!$form_state['rebuild']) {
    return;
  }

  $section_inputs = [];
  // phpcs:ignore DrupalPractice.General.FormStateInput -- 'input' needs to be used here.
  if (isset($form_state['input']['rules']['group-' . AC_ATTR_ADMIN_SECTIONS_RGID]['rules']['list'])) {
    // phpcs:ignore DrupalPractice.General.FormStateInput -- 'input' needs to be used here.
    $section_inputs = &$form_state['input']['rules']['group-' . AC_ATTR_ADMIN_SECTIONS_RGID]['rules']['list'];
  }

  // Was there already new sections added? If so, restore them.
  foreach ($section_inputs as $section_id => $section) {
    $restored_rule = _ac_attr_admin_get_new_rule_data(AC_ATTR_ADMIN_SECTIONS_RGID);

    // New rule?
    if (strpos($section_id, 'new-') === 0) {
      $restored_rule['new_rule'] = TRUE;
      $restored_rule['new_rule_id'] = $section_id;
      $next_new_rule_id = ((int) substr($section_id, 4)) + 1;
    }
    else {
      $restored_rule['new_rule'] = FALSE;
      $restored_rule['rid'] = (int) substr($section_id, 4);
    }

    if (isset($section['field_path_patterns'])) {
      $restored_rule['field_path_patterns'] = $section['field_path_patterns'];
    }
    elseif (isset($rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['field_path_patterns'])) {
      $restored_rule['field_path_patterns'] = $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['field_path_patterns'];
    }

    if (isset($section['value'])) {
      $restored_rule['value'] = $section['value'];
    }
    elseif (isset(
      $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['value']
    )) {
      $restored_rule['value'] = $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['value'];
    }

    if (isset($section['disabled'])) {
      $restored_rule['disabled'] = $section['disabled'];
    }
    elseif (isset($rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['disabled'])) {
      $restored_rule['disabled'] = $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['disabled'];
    }

    if (isset($section['title'])) {
      $restored_rule['title'] = $section['title'];
    }
    elseif (isset($rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['title'])) {
      $restored_rule['title'] = $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]['title'];
    }

    $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id] = $restored_rule;
  }

  // Were any db saved section removed? Make sure they're not restored.
  $section_ids = array_keys($section_inputs);
  foreach ($rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID] as $section_id => $section) {
    if (!in_array($section_id, $section_ids)) {
      unset($rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$section_id]);
    }
  }

  // Was the add section button clicked?
  if ($form_state['triggering_element']['#name'] === 'add_section') {
    $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID]['new-' . $next_new_rule_id]
      = _ac_attr_admin_get_new_rule_data(AC_ATTR_ADMIN_SECTIONS_RGID);
  }

  // Was a remove section button clicked?
  if (strpos($form_state['triggering_element']['#name'], 'delete-section-') === 0) {
    $index_to_remove = substr($form_state['triggering_element']['#name'], strlen('delete-section-'));
    if (isset(
      $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$index_to_remove]
    )) {
      unset(
        $rules_by_group_id[AC_ATTR_ADMIN_SECTIONS_RGID][$index_to_remove]
      );
    }
  }
}

/**
 * Returns a form element fieldset for a nested group of rule tokens.
 *
 * @return array
 *   A form element.
 */
function _ac_attr_admin_rules_get_nested_group_form_element($group_name = '') {
  return [
    '#type' => 'fieldset',
    '#title' => empty($group_name)
    ? t('Nested Rule Group')
    : t('Nested Rule Group: @group_name', ['@group_name' => $group_name]),
    '#collapsible' => TRUE,
    '#weight' => 10,
  ];
}

/**
 * Submit handler for ac_attr_admin_rules.
 */
function ac_attr_admin_rules_submit($form, &$form_state) {
  variable_set(
    'ac_attr_root_field_path_patterns',
    implode("\n",
      array_filter(
        array_map(
          'trim',
          preg_split(
            "/\s+/",
            $form_state['values']['root_patterns']
          )
        ),
        function ($item) {
          return !empty($item);
        }
      )
    )
  );

  $rule_group_stack = [];
  foreach ($form_state['values']['rules'] as $rule_group) {
    $rule_group_stack[] = $rule_group;
  }

  // The rule ids already in the database.
  $db_rule_ids = array_reduce(
    db_select('ac_attr_rules', 'r')
      ->fields('r', ['rid'])
      ->execute()
      ->fetchAll(PDO::FETCH_ASSOC),
    function ($carry, $rule) {
      $carry[] = $rule['rid'];
      return $carry;
    },
    []
  );

  $rule_ids_returned = [];

  while (count($rule_group_stack) > 0) {
    $rule_group = array_pop($rule_group_stack);

    // Save all the rule changes.
    if (array_key_exists('list', $rule_group['rules'])) {
      foreach ($rule_group['rules']['list'] as $rule) {
        // New Rule?
        if (strpos($rule['rid'], 'new-') === 0) {
          // Insert.
          db_insert('ac_attr_rules')
            ->fields([
              'rgid' => (int) $rule_group['rgid'],
              'field_path_patterns' => implode("\n",
                array_filter(
                  array_map('trim', preg_split("/\s+/", $rule['patterns'])),
                  function ($item) {
                    return !empty($item);
                  }
                )
              ),
              'value' => $rule['value'],
              'requires_subvalue' => (int) $rule['requires_subvalue'],
              'title' => $rule['title'],
              'disabled' => (int) $rule['disabled'],
              'built_in' => (int) FALSE,
            ])
            ->execute();
        }
        // Existing Rule?
        elseif (strpos($rule['rid'], 'rid-') === 0) {
          $rid = (int) substr($rule['rid'], 4);

          $rule_ids_returned[] = $rid;

          $fields = [
            'field_path_patterns' => implode("\n",
              array_filter(
                array_map('trim', preg_split("/\s+/", $rule['patterns'])),
                function ($item) {
                  return !empty($item);
                }
              )
            ),
            'disabled' => (int) $rule['disabled'],
          ];

          if (!$rule['built_in']) {
            $fields += [
              'value' => $rule['value'],
              'title' => $rule['title'],
            ];
          }

          db_update('ac_attr_rules')
            ->fields($fields)
            ->condition('rid', $rid)
            ->execute();
        }
      }
    }

    // Add all the subrules to the stack.
    if (array_key_exists('subrules', $rule_group)) {
      foreach ($rule_group['subrules'] as $subrule_group) {
        $rule_group_stack[] = $subrule_group;
      }
    }
  }

  $rule_ids_to_remove = array_reduce(
    $db_rule_ids,
    function ($carry, $db_rid) use ($rule_ids_returned) {
      if (!in_array($db_rid, $rule_ids_returned)) {
        $carry[] = $db_rid;
      }
      return $carry;
    },
    []
  );

  if (count($rule_ids_to_remove) > 0) {
    // Delete the removed rules.
    db_delete('ac_attr_rules')
      ->condition('rid', $rule_ids_to_remove)
      ->execute();
  }

  // Save the rules to the cache.
  _ac_attr_cache_rules();
}

/**
 * Validation handler for ac_attr_admin_rules.
 */
function ac_attr_admin_rules_validate($form, &$form_state) {
  foreach ($form_state['values']['rules'] as $group_key => $group) {
    if (isset($group['rules']['list'])) {
      foreach ($group['rules']['list'] as $rule_key => $rule) {
        // Sections?
        if ($group['rgid'] === AC_ATTR_ADMIN_SECTIONS_RGID) {
          // Section values must start with "section-".
          if (strpos($rule['value'], 'section-') !== 0) {
            form_set_error(
              "rules][$group_key][rules][list][$rule_key][value",
              t('Section values must start with "section-", "@given" was give.', [
                '@given' => $rule['value'],
              ])
            );
          }
          // Something must come after "section-".
          elseif (strlen($rule['value']) === strlen('section-')) {
            form_set_error(
              "rules][$group_key][rules][list][$rule_key][value",
              t('No section name given after "section-".')
            );
          }
        }
      }
    }
  }
}

/**
 * Overrides Form.
 */
function ac_attr_admin_overrides($form, &$form_state) {
  drupal_add_js(drupal_get_path('module', 'ac_attr') . '/js/admin_forms.js');
  drupal_add_css(drupal_get_path('module', 'ac_attr') . '/css/admin_forms.css');

  $about = t('If you want to override the autocomplete attribute of specific fields on specific forms you can do so here.<br><br>To set generic rules, use the <a href="@rules_url">rule page</a>.<br>', [
    '@rules_url' => 'admin/config/content/ac_attr/rules',
  ]);

  $form['about'] = [
    '#type' => 'markup',
    '#markup' => "<p>$about<p>",
    '#prefix' => '<div class="messages info">',
    '#suffix' => '</div>',
    '#weight' => -10,
  ];

  _ac_attr_admin_add_form_helpers($form);

  $form['#attributes'] = [
    'class' => ['ac-attr-admin'],
  ];

  // Get overrides from the database.
  $overrides = array_reduce(
    db_select('ac_attr_overrides', 'o')
      ->fields('o')
      ->orderBy('form_id', 'ASC')
      ->execute()
      ->fetchAll(PDO::FETCH_ASSOC),
    function ($carry, $override) {
      return $carry + ['oid-' . $override['oid'] => $override];
    },
    []
  );

  $next_new_rule_id = 0;
  _ac_attr_admin_overrides_rebuild($form_state, $overrides, $next_new_rule_id);

  $form['overrides'] = [
    '#tree' => TRUE,
    '#weight' => 5,
  ];

  // Overrides?
  if (count($overrides) > 0) {
    foreach ($overrides as $override) {
      $new_override = isset($override['new_override']) && $override['new_override'];

      $completely_new_rule = $new_override && empty($override['value']);

      $override_id = $new_override
        ? (isset($override['new_override_id'])
          ? $override['new_override_id']
          : ('new-' . ($next_new_rule_id++)))
        : 'oid-' . $override['oid'];

      $fieldset_title = $completely_new_rule
        ? t('New Override')
        : (empty($override['title'])
          ? ($override['disabled']
            ? t('Override for @form_id @field_path (@ac_value) [disabled]', [
              '@form_id' => $override['form_id'],
              '@field_path' => $override['field_path'],
              '@ac_value' => $override['value'],
            ])
            : t('Override for @form_id @field_path (@ac_value)', [
              '@form_id' => $override['form_id'],
              '@field_path' => $override['field_path'],
              '@ac_value' => $override['value'],
            ])
          )
          : ($override['disabled']
            ? t('@title [disabled]', ['@title' => $override['title']])
            : $override['title']
          )
        );

      $form['overrides'][$override_id] = [
        '#type' => 'fieldset',
        '#title' => $fieldset_title,
        '#collapsible' => TRUE,
        '#collapsed' => !$completely_new_rule,
      ];
      $form['overrides'][$override_id]['oid'] = [
        '#type' => 'value',
        '#value' => $override_id,
      ];
      $form['overrides'][$override_id]['title'] = [
        '#type' => 'textfield',
        '#title' => t('Title'),
        '#default_value' => $override['title'],
        '#attributes' => [
          'placeholder' => 'My Override',
          'autocomplete' => 'off',
        ],
        '#required' => FALSE,
        '#description' => t('The title that is displayed for this override.'),
      ];
      $form['overrides'][$override_id]['form_id'] = [
        '#type' => 'textfield',
        '#title' => t('Form ID'),
        '#default_value' => $override['form_id'],
        '#attributes' => [
          'placeholder' => 'my_form',
          'class' => ['form-ids'],
          'autocomplete' => 'off',
        ],
        '#required' => TRUE,
        '#description' => t(
          'The ID of the form that this override applies to.'
        ),
      ];
      $form['overrides'][$override_id]['field_path'] = [
        '#type' => 'textfield',
        '#title' => t('Field Path'),
        '#default_value' => $override['field_path'],
        '#attributes' => [
          'placeholder' => 'foo.bar.baz',
          'class' => ['field-path'],
          'autocomplete' => 'off',
        ],
        '#required' => TRUE,
        '#description' => t(
          "The field path for the field that this override is applied to."
        ),
      ];
      $form['overrides'][$override_id]['value'] = [
        '#type' => 'textfield',
        '#title' => t('Token(s)'),
        '#default_value' => $override['value'],
        '#attributes' => [
          'placeholder' => 'foo',
          'autocomplete' => 'off',
        ],
        '#required' => TRUE,
        '#description' => t(
          'The value that will be set on the autocomplete attribute.'
        ),
      ];
      $form['overrides'][$override_id]['disabled'] = [
        '#type' => 'checkbox',
        '#title' => t('Disbaled'),
        '#attributes' => [
          'autocomplete' => 'off',
        ],
        '#default_value' => $override['disabled'],
        '#description' => t('When set, this override will never be applied.'),
      ];
      // Add a button to remove this override.
      $form['overrides'][$override_id]['actions'] = [
        '#type' => 'actions',
        'delete' => [
          '#type' => 'button',
          '#value' => t('Remove'),
          '#name' => "delete-override-$override_id",
          '#attributes' => [
            'formnovalidate' => TRUE,
          ],
          '#limit_validation_errors' => [],
          // TODO: Add AJAX handler.
          '#weight' => 2,
        ],
        '#weight' => 100,
      ];
    }
  }
  // No overrides?
  else {
    $form['overrides']['empty'] = [
      '#type' => 'markup',
      '#prefix' => '<div class="no-overrides">',
      '#markup' => t('There are currently no overrides in effect.'),
      '#suffix' => '</div>',
      '#weight' => 5,
    ];
  }

  // Add the action buttons.
  $form['actions'] = [
    '#type' => 'actions',
    'add' => [
      '#type' => 'button',
      '#value' => t('Add Override'),
      '#name' => 'add_override',
      // TODO: Add AJAX handler.
    ],
    'submit' => [
      '#type' => 'submit',
      '#value' => t('Submit'),
    ],
  ];

  $form['#validate'][] = 'ac_attr_admin_overrides_validate';

  return $form;
}

/**
 * Restore the ac_attr_admin_overrides form if it is being rebuild.
 *
 * @param array $form_state
 *   The form state.
 * @param array $overrides
 *   The overrides to apply.
 * @param int $next_new_rule_id
 *   The next new rule id to use.
 */
function _ac_attr_admin_overrides_rebuild(
  array &$form_state,
  array &$overrides,
  &$next_new_rule_id
) {
  // If the form is not being rebuilt, get out of here.
  if (!$form_state['rebuild']) {
    return;
  }

  $override_inputs = [];
  // phpcs:ignore DrupalPractice.General.FormStateInput -- 'input' needs to be used here.
  if (isset($form_state['input']['overrides'])) {
    // phpcs:ignore DrupalPractice.General.FormStateInput -- 'input' needs to be used here.
    $override_inputs = &$form_state['input']['overrides'];
  }

  // Was there already new overrides added? If so, restore them.
  foreach ($override_inputs as $override_id => $override) {
    $restored_override = _ac_attr_admin_get_new_rule_override_data();

    // New override?
    if (strpos($override_id, 'new-') === 0) {
      $restored_override['new_override'] = TRUE;
      $restored_override['new_override_id'] = $override_id;
      $next_new_rule_id = ((int) substr($override_id, 4)) + 1;
    }
    else {
      $restored_override['new_override'] = FALSE;
      $restored_override['oid'] = (int) substr($override_id, 4);
    }

    if (isset($override['form_id'])) {
      $restored_override['form_id'] = $override['form_id'];
    }
    elseif (isset($overrides[$override_id]['form_id'])) {
      $restored_override['form_id'] = $overrides[$override_id]['form_id'];
    }

    if (isset($override['field_path'])) {
      $restored_override['field_path'] = $override['field_path'];
    }
    elseif (isset($overrides[$override_id]['field_path'])) {
      $restored_override['field_path'] = $overrides[$override_id]['field_path'];
    }

    if (isset($override['value'])) {
      $restored_override['value'] = $override['value'];
    }
    elseif (isset($overrides[$override_id]['value'])) {
      $restored_override['value'] = $overrides[$override_id]['value'];
    }

    if (isset($override['disabled'])) {
      $restored_override['disabled'] = $override['disabled'];
    }
    elseif (isset($overrides[$override_id]['disabled'])) {
      $restored_override['disabled'] = $overrides[$override_id]['disabled'];
    }

    if (isset($override['title'])) {
      $restored_override['title'] = $override['title'];
    }
    elseif (isset($overrides[$override_id]['title'])) {
      $restored_override['title'] = $overrides[$override_id]['title'];
    }

    $overrides[$override_id] = $restored_override;
  }

  // Were any db saved override removed? Make sure they're not restored.
  $override_ids = array_keys($override_inputs);
  foreach ($overrides as $override_id => $override) {
    if (!in_array($override_id, $override_ids)) {
      unset($overrides[$override_id]);
    }
  }

  // Was the add override button clicked?
  if ($form_state['triggering_element']['#name'] === 'add_override') {
    $overrides[] = _ac_attr_admin_get_new_rule_override_data();
  }

  // Was a remove button clicked?
  if (strpos($form_state['triggering_element']['#name'], 'delete-override-') === 0) {
    $index_to_remove = substr($form_state['triggering_element']['#name'], strlen('delete-override-'));
    if (isset($overrides[$index_to_remove])) {
      unset($overrides[$index_to_remove]);
    }
  }
}

/**
 * Submit handler for Overrides forms.
 */
function ac_attr_admin_overrides_submit($form, &$form_state) {
  // The override ids already in the database.
  $db_override_ids = array_reduce(
    db_select('ac_attr_overrides', 'o')
      ->fields('o', ['oid'])
      ->execute()
      ->fetchAll(PDO::FETCH_ASSOC),
    function ($carry, $override) {
      $carry[] = $override['oid'];
      return $carry;
    },
    []
  );

  $override_ids_returned = [];

  if (isset($form_state['values']['overrides'])) {
    foreach ($form_state['values']['overrides'] as $override) {
      // New Override?
      if (strpos($override['oid'], 'new-') === 0) {
        // Insert.
        db_insert('ac_attr_overrides')
          ->fields([
            'title' => $override['title'],
            'form_id' => $override['form_id'],
            'field_path' => $override['field_path'],
            'value' => $override['value'],
            'disabled' => (int) $override['disabled'],
          ])
          ->execute();
      }
      // Existing Override?
      elseif (strpos($override['oid'], 'oid-') === 0) {
        $oid = (int) substr($override['oid'], 4);

        $override_ids_returned[] = $oid;

        $fields = [
          'title' => $override['title'],
          'form_id' => $override['form_id'],
          'field_path' => $override['field_path'],
          'value' => $override['value'],
          'disabled' => (int) $override['disabled'],
        ];

        db_update('ac_attr_overrides')
          ->fields($fields)
          ->condition('oid', $oid)
          ->execute();
      }
    }
  }

  $override_ids_to_remove = array_reduce(
    $db_override_ids,
    function ($carry, $db_rid) use ($override_ids_returned) {
      if (!in_array($db_rid, $override_ids_returned)) {
        $carry[] = $db_rid;
      }
      return $carry;
    },
    []
  );

  if (count($override_ids_to_remove) > 0) {
    // Delete the removed overrides.
    db_delete('ac_attr_overrides')
      ->condition('oid', $override_ids_to_remove)
      ->execute();
  }
}

/**
 * Validation handler for ac_attr_admin_overrides.
 */
function ac_attr_admin_overrides_validate($form, &$form_state) {
  if (isset($form_state['values']['overrides'])) {
    $override_conflicts = [];
    foreach ($form_state['values']['overrides'] as $id => $override) {
      // Already overridden?
      if (isset($override_conflicts[$override['form_id']][$override['field_path']])) {
        form_set_error(
          "overrides][$id",
          t(
            'Cannot add multiple overrides for the same field (form: "@form_id", field path: "@field_path").',
            [
              '@form_id' => $override['form_id'],
              '@field_path' => $override['field_path'],
            ]
          )
        );
      }

      $override_conflicts[$override['form_id']][$override['field_path']] = TRUE;
    }
  }
}

/**
 * Settings Form.
 */
function ac_attr_admin_settings($form, &$form_state) {
  drupal_add_js(drupal_get_path('module', 'ac_attr') . '/js/admin_forms.js');
  drupal_add_css(drupal_get_path('module', 'ac_attr') . '/css/admin_forms.css');

  _ac_attr_admin_add_form_helpers($form);

  $form['#attributes'] = [
    'class' => ['ac-attr-admin'],
  ];

  $disabled_forms = variable_get(
    'ac_attr_disabled_forms',
    AC_ATTR_DEFAULT_DISABLED_FORMS
  );

  $form['disabled_forms'] = [
    '#type' => 'textarea',
    '#title' => t('Disabled For These Forms'),
    '#default_value' => $disabled_forms,
    '#attributes' => [
      'class' => ['field-disabled-forms'],
      'autocomplete' => 'off',
    ],
    '#rows' => max(5, min(10,
      substr_count($disabled_forms, "\n") + 1
    )),
    '#weight' => 10,
    '#description' => t(
      "This module will not effect any forms with the form IDs listed here."
    ),
  ];

  // Add the action buttons.
  $form['actions'] = [
    '#type' => 'actions',
    'submit' => [
      '#type' => 'submit',
      '#value' => t('Submit'),
    ],
  ];

  return $form;
}

/**
 * Submit handler for Settings forms.
 */
function ac_attr_admin_settings_submit($form, &$form_state) {
  variable_set(
    'ac_attr_disabled_forms',
    implode("\n",
      array_filter(
        array_map(
          'trim',
          preg_split(
            "/\s+/",
            $form_state['values']['disabled_forms']
          )
        ),
        function ($item) {
          return !empty($item);
        }
      )
    )
  );
}

/**
 * The rules flattend out and rendered as a table.
 */
function ac_attr_admin_flattened_rules() {
  drupal_add_css(
    drupal_get_path('module', 'ac_attr') . '/css/admin_flattened_rules.css'
  );

  $flattened_rules_table_header = [
    t('Field Path Pattern'),
    t('Autocomplete Attribute Value'),
  ];

  $flattened_rules_table_rows = [];
  $flattened_rules = _ac_attr_get_flattened_rules();
  foreach ($flattened_rules as $rule) {
    $flattened_rules_table_rows[] = [
      ['data' => $rule['pattern'], 'class' => 'pattern'],
      ['data' => $rule['value'], 'class' => 'value'],
    ];
  }

  return [
    '#type' => 'markup',
    '#markup' => t('<div class="messages info"><h1>Flattened Rules</h1><p>The following list is of all the rules as defined on the <a href="@rules_url">ac_attr\'s rule page</a>, flattened out into a list that displays the corresponding autocomplete attribute value.</p><p>Please note that disabled rules will not be listed here.</p></div><div class="stats info"><span class="label">Total number of rules in effect:</span> <span class="value">@rules_in_effect</span></div>', [
      '@rules_url' => url('admin/config/content/ac_attr/rules'),
      '@rules_in_effect' => count($flattened_rules_table_rows),
    ])
    . theme('table', [
      'header' => $flattened_rules_table_header,
      'rows' => $flattened_rules_table_rows,
      'attributes' => [
        'class' => ['flattened-rules'],
      ],
      'empty' => t('No rules in effect.'),
    ]),
  ];
}

/**
 * Get a new rule.
 *
 * @param int $rgid
 *   The rule group the rule should be in.
 *
 * @return array
 *   Data for a new rule.
 */
function _ac_attr_admin_get_new_rule_data($rgid) {
  return [
    'rid' => NULL,
    'rgid' => $rgid,
    'field_path_patterns' => '',
    'value' => '',
    'built_in' => FALSE,
    'requires_subvalue' => TRUE,
    'disabled' => FALSE,
    'title' => '',
    'description' => '',
    'new_rule' => TRUE,
  ];
}

/**
 * Get a new rule override.
 *
 * @return array
 *   Data for a new rule override.
 */
function _ac_attr_admin_get_new_rule_override_data() {
  return [
    'oid' => NULL,
    'title' => '',
    'form_id' => '',
    'field_path' => '',
    'value' => '',
    'disabled' => FALSE,
    'new_override' => TRUE,
  ];
}

/**
 * Add form helpers to the given form.
 *
 * @param array $form
 *   The form to add the helpers to.
 */
function _ac_attr_admin_add_form_helpers(array &$form) {
  // These are only displayed if JS is enabled, so hide them and let JS show
  // them.
  $form['helpers']['#prefix'] = '<div class="helpers form-wrapper" style="display:none;">';
  $form['helpers']['#suffix'] = '</div>';
  $form['helpers']['#weight'] = -1;

  // Add the collapse/expand button.
  $form['helpers']['collapse_expand'] = [
    '#type' => 'button',
    '#attributes' => [
      'class' => ['collapse-expand'],
    ],
  ];
}
